/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#include <config.h>
#include <assert.h>
#include "common.h"
#include "approx.h"
#include "error.h"
#include "misc.h"
#include "tokens.h"
#include "var.h"

/* Specifies how to make an aggregate variable. */
typedef struct agr_var_struct agr_var;
struct agr_var_struct
  {
    agr_var *next;		/* Next in list. */

    /* Collected during parsing. */
    variable *src;		/* Source variable. */
    variable *dest;		/* Target variable. */
    int function;		/* Function. */
    int include_missing;	/* 1=Include user-missing values. */
    value arg[2];		/* Arguments. */

    /* Accumulated during AGGREGATE execution. */
    double dbl[2];
    int int1, int2;
    char *string;
  };

/* Aggregation functions. */
enum
  {
    NONE, SUM, MEAN, SD, MAX, MIN, PGT, PLT, PIN, POUT, FGT, FLT, FIN,
    FOUT, N, NU, NMISS, NUMISS, FIRST, LAST,
    N_AGR_FUNCS, N_NO_VARS, NU_NO_VARS,
    FSTRING = 1<<5, /* String function bit. */
    FWEIGHT = 1<<6, /* Weighted function bit. */
    FOPTIONS = FSTRING | FWEIGHT, /* Function options mask. */
  };

/* Attributes of an aggregation function. */
typedef struct
  {
    char *name;		/* Aggregation function name. */
    int n_args;		/* Number of arguments. */
    int alpha_type;	/* When given ALPHA arguments, output type. */
    fmt_spec format;	/* Format specification if alpha_type != ALPHA. */
  }
agr_func;

/* Attributes of aggregation functions. */
static agr_func agr_func_tab[] = 
  {
    {"<NONE>",  0, -1,      {0, 0, 0}},
    {"SUM",     0, -1,      {FMT_F, 8, 2}},
    {"SD",      0, -1,      {FMT_F, 8, 2}},
    {"MAX",     0, ALPHA,   {-1, -1, -1}}, 
    {"MIN",     0, ALPHA,   {-1, -1, -1}}, 
    {"PGT",     1, NUMERIC, {FMT_F, 5, 1}},      
    {"PLT",     1, NUMERIC, {FMT_F, 5, 1}},       
    {"PIN",     2, NUMERIC, {FMT_F, 5, 1}},       
    {"POUT",    2, NUMERIC, {FMT_F, 5, 1}},       
    {"FGT",     1, NUMERIC, {FMT_F, 5, 3}},       
    {"FLT",     1, NUMERIC, {FMT_F, 5, 3}},       
    {"FIN",     2, NUMERIC, {FMT_F, 5, 3}},       
    {"FOUT",    2, NUMERIC, {FMT_F, 5, 3}},       
    {"N",       0, NUMERIC, {FMT_F, 7, 0}},       
    {"NU",      0, NUMERIC, {FMT_F, 7, 0}},       
    {"NMISS",   0, NUMERIC, {FMT_F, 7, 0}},       
    {"NUMISS",  0, NUMERIC, {FMT_F, 7, 0}},       
    {"FIRST",   0, ALPHA,   {-1, -1, -1}}, 
    {"LAST",    0, ALPHA,   {-1, -1, -1}},
    {NULL,      0, -1,      {-1, -1, -1}},
    {"N",       0, NUMERIC, {FMT_F, 7, 0}},
    {"NU",      0, NUMERIC, {FMT_F, 7, 0}},
  };

/* Output file, or NULL for the active file. */
static file_handle *outfile;

/* Missing value types. */
enum
  {
    ITEMWISE,		/* Missing values item by item. */
    COLUMNWISE,		/* Missing values column by column. */
  };

/* ITEMWISE or COLUMNWISE. */
static int missing;

/* Aggregate variables. */
static agr_var *agr_first, *agr_next;

/* Aggregate dictionary. */
static dictionary *agr_dict;

/* Number of cases passed through aggregation. */
static int case_count;

static void initialize_aggregate_functions (void);

/* Prototypes. */
static int parse_aggregate_functions (void);
static void free_aggregate_functions (void);
#if DEBUGGING
static void debug_print (int flags);
#endif

/* Parsing. */

/* Parses and executes the AGGREGATE procedure. */
int
cmd_aggregate (void)
{
  /* From sort.c. */
  int parse_sort_variables (void);
  
  /* Have we seen these subcommands? */
  unsigned seen = 0;

  outfile = NULL;
  missing = ITEMWISE;
  v_sort = NULL;
  
  agr_dict = new_dictionary ();
  
  match_id (AGGREGATE);

  /* Read most of the subcommands. */
  for (;;)
    {
      match_tok('/');
      
      if (match_id (OUTFILE))
	{
	  if (seen & 1)
	    {
	      free (v_sort);
	      free_dictionary (agr_dict);
	      return msg (SE, _("OUTFILE specified multiple times."));
	    }
	  seen |= 1;
	      
	  match_tok ('=');
	  if (match_tok ('*'))
	    outfile = NULL;
	  else 
	    {
	      outfile = fh_parse_file_handle ();
	      if (outfile == NULL)
		{
		  free (v_sort);
		  free_dictionary (agr_dict);
		  return 0;
		}
	    }
	}
      else if (match_id (MISSING))
	{
	  match_tok ('=');
	  if (!match_id (COLUMNWISE))
	    {
	      free (v_sort);
	      free_dictionary (agr_dict);
	      return syntax_error (_("while expecting COLUMNWISE"));
	    }
	  missing = COLUMNWISE;
	}
      else if (match_id (DOCUMENT))
	seen |= 2;
      else if (match_id (PRESORTED))
	seen |= 4;
      else if (match_id (BREAK))
	{
	  if (seen & 8)
	    {
	      free (v_sort);
	      free_dictionary (agr_dict);
	      return msg (SE, _("BREAK specified multiple times."));
	    }
	  seen |= 8;

	  match_tok ('=');
	  if (!parse_sort_variables ())
	    {
	      free_dictionary (agr_dict);
	      return 0;
	    }
	  
	  {
	    int i;
	    
	    for (i = 0; i < nv_sort; i++)
	      {
		variable *v;
	      
		v = dup_variable (agr_dict, v_sort[i], v_sort[i]->name);
		assert (v);
	      }
	  }
	}
      else break;
    }

  /* Check for proper syntax. */
  if (!(seen & 8))
    msg (SW, _("BREAK subcommand not specified."));
      
  /* Read in the aggregate functions. */
  if (!parse_aggregate_functions ())
    {
      free_aggregate_functions ();
      free (v_sort);
      free_dictionary (agr_dict);
      return 0;
    }

#if DEBUGGING
  debug_print (seen);
#endif

  /* Initialize. */
  case_count = 0;
  initialize_aggregate_functions ();

  /* How to implement all this... There are three important variables:
     whether output is going to the active file (0) or a separate file
     (1); whether the input data is presorted (0) or needs sorting
     (1); whether there is a temporary transformation (1) or not (0).
     The eight cases are as follows:

     000 (0): pass it through an aggregate transformation that
     modifies the data.

     001 (1): cancel the temporary transformation and handle as 000.

     010 (2): set up a SORT CASES and aggregate the output, writing
     the results to the active file.
     
     011 (3): cancel the temporary transformation and handle as 010.

     100 (4): pass it through an aggregate transformation that doesn't
     modify the data but merely writes it to the output file.

     101 (5): handled as 100.

     110 (6): set up a SORT CASES and capture the output, aggregate
     it, write it to the output file without modifying the active
     file.

     111 (7): handled as 110. */
  
  {
    unsigned type = 0;

    if (outfile != NULL)
      type |= 4;
    if (nv_sort != 0 && (seen & 4) == 0)
      type |= 2;
    if (temporary)
      type |= 1;

    switch (type)
      {
      case 1:
	cancel_temporary ();
	/* fall through */
      case 0:
	break;

      case 3:
	cancel_temporary ();
	/* fall through */
      case 2:
	break;

      case 4:
      case 5:
	break;

      case 6:
      case 7:
	break;

      default:
	assert (0);
      }
  }
  
  /* Clean up. */
  free (v_sort);
  free_aggregate_functions ();
  free_dictionary (agr_dict);
  
  return 1;
}

/* Parse all the aggregate functions. */
static int
parse_aggregate_functions (void)
{
  agr_first = agr_next = NULL;

  /* Anticipate weighting for optimization later. */
  update_weighting (&default_dict);
  
  /* Parse everything. */
  for (;;)
    {
      char **dest;
      char **dest_label;
      int n_dest;

      int include_missing;
      agr_func *function;
      int func_index;

      value arg[2];

      variable **src;
      int n_src;

      int i;

      dest = NULL;
      dest_label = NULL;
      n_dest = 0;
      src = NULL;
      n_src = 0;
      arg[0].c = NULL;
      arg[1].c = NULL;

      /* Parse the list of target variables. */
      while (!match_tok ('='))
	{
	  int n_dest_prev = n_dest;
	  
	  if (!parse_DATA_LIST_vars (&dest, &n_dest, PV_APPEND | PV_SINGLE))
	    goto lossage;

	  /* Assign empty labels. */
	  {
	    int i;

	    dest_label = xrealloc (dest_label, sizeof (char *) * n_dest);
	    for (i = n_dest_prev; i < n_dest; i++)
	      dest_label[i] = NULL;
	  }
	  
	  if (token == STRING)
	    {
	      if (strlen (tokstr) > 120)
		tokstr[120] = 0;
	      dest_label[n_dest - 1] = xstrdup (tokstr);
	      get_token ();
	    }
	}

      /* Get the name of the aggregation function. */
      if (token != ID)
	{
	  syntax_error (_("expecting aggregation function"));
	  goto lossage;
	}

      include_missing = 0;
      if (tokstr[strlen(tokstr) - 1] == '.')
	{
	  include_missing = 1;
	  tokstr[strlen(tokstr) - 1] == '.';
	}
      
      for (function = agr_func_tab; function->name; function++)
	if (streq (function->name, tokstr))
	  break;
      if (NULL == function->name)
	{
	  msg (SE, _("Unknown aggregation function %s."), tokstr);
	  goto lossage;
	}
      func_index = function - agr_func_tab;
      get_token ();

      /* Leading lparen. */
      if (!match_tok ('('))
	{
	  syntax_error (_("expecting `('"));
	  goto lossage;
	}

      /* Parse list of source variables. */
      {
	pv_opts options = PV_NONE;

	if (func_index == SUM || func_index == MEAN || func_index == SD)
	  options |= PV_NUMERIC;
	else if (function->n_args)
	  options |= PV_SAME_TYPE;

	if (!parse_variables (&default_dict, &src, &n_src, options))
	  goto lossage;
      }

      /* Parse function arguments, for those functions that require
         arguments. */
      if (function->n_args != 0)
	for (i = 0; i < function->n_args; i++)
	  {
	    int type;
	    
	    match_tok (',');
	    if (token == STRING)
	      {
		arg[i].c = xstrdup (tokstr);
		type = ALPHA;
	      }
	    else if (token == NUM)
	      {
		arg[i].f = tokval;
		type = NUMERIC;
	      } else {
		msg (SE, _("Missing argument %d to %s."), i + 1, function->name);
		goto lossage;
	      }
	    
	    get_token ();

	    if (type != src[0]->type)
	      {
		msg (SE, _("Arguments to %s must be of same type as "
			   "source variables."),
		     function->name);
		goto lossage;
	      }
	  }

      /* Trailing rparen. */
      if (!match_tok(')'))
	{
	  syntax_error (_("expecting `)'"));
	  goto lossage;
	}

      /* Now check that the number of source variables match the
	 number of target variables.  Do this here because if we do it
	 earlier then the user can get very misleading error messages;
	 i.e., `AGGREGATE x=SUM(y t).' will get this error message
	 when a proper message would be more like `unknown variable
	 t'. */
      if (n_src != n_dest)
	{
	  msg (SE, _("Number of source variables (%d) does not match "
		     "number of target variables (%d)."),
	       n_src, n_dest);
	  goto lossage;
	}
	
      /* Finally add these to the linked list of aggregation
         variables. */
      for (i = 0; i < n_dest; i++)
	{
	  agr_var *v = xmalloc (sizeof (agr_var));

	  /* Add variable to chain. */
	  if (agr_first)
	    agr_next = agr_next->next = v;
	  else
	    agr_first = agr_next = v;
	  agr_next->next = NULL;
      
	  agr_next->src = src[i];

	  /* Create the target variable in the aggregate
             dictionary. */
	  {
	    int output_type;
	    variable *destvar;

	    if (agr_next->src->type == NUMERIC)
	      output_type = NUMERIC;
	    else
	      output_type = function->alpha_type;

	    if (function->alpha_type == ALPHA)
	      destvar = dup_variable (agr_dict, agr_next->src, dest[i]);
	    else
	      {
		destvar = create_variable (agr_dict, dest[i], output_type,
					   agr_next->src->width);
		if (output_type == NUMERIC)
		  destvar->print = destvar->write = function->format;
		if (output_type == NUMERIC && default_dict.weight_index != -1
		    && (func_index == N || func_index == N_NO_VARS
			|| func_index == NU || func_index == NU_NO_VARS))
		  {
		    fmt_spec f = {FMT_F, 8, 2};
		      
		    destvar->print = destvar->write = f;
		  }
	      }

	    if (!destvar)
	      {
		msg (SE, _("Variable name %s is not unique within the "
			   "aggregate file dictionary, which contains "
			   "the aggregate variables and the break "
			   "variables."),
		     dest[i]);
		free (dest[i]);
		goto lossage;
	      }

	    free (dest[i]);

	    if (dest_label[i])
	      {
		destvar->label = dest_label[i];
		dest_label[i] = NULL;
	      }
	    else if (function->alpha_type == ALPHA)
	      destvar->print = destvar->write = function->format;
	  }
	  
	  agr_next->function = func_index;
	  if (src[i]->type == ALPHA)
	    agr_next->function |= FSTRING;
	  if (default_dict.weight_index != -1)
	    agr_next->function |= FWEIGHT;
	      
	  agr_next->include_missing = include_missing;
	  
	  {
	    int j;

	    if (agr_next->src->type == NUMERIC)
	      for (j = 0; j < function->n_args; j++)
		agr_next->arg[j].f = arg[j].f;
	    else
	      for (j = 0; j < function->n_args; j++)
		agr_next->arg[j].c = xstrdup (arg[j].c);
	  }
	}
      
      if (src[0]->type == ALPHA)
	for (i = 0; i < function->n_args; i++)
	  {
	    free (arg[i].c);
	    arg[i].c = NULL;
	  }

      free (src);
      free (dest);
      free (dest_label);

      if (!match_id ('/'))
	{
	  if (token == '.')
	    return 1;
	  else
	    return syntax_error ("expecting end of command");
	}
      continue;
      
    lossage:
      for (i = 0; i < n_dest; i++)
	{
	  free (dest[i]);
	  free (dest_label[i]);
	}
      free (dest);
      free (dest_label);
      free (arg[0].c);
      free (arg[1].c);
      if (src && n_src && src[0]->type == ALPHA)
	for (i = 0; i < function->n_args; i++)
	  {
	    free(arg[i].c);
	    arg[i].c = NULL;
	  }
      free (src);
	
      return 0;
    }
}

/* Frees all the state for the AGGREGATE procedure. */
static void
free_aggregate_functions (void)
{
  agr_var *iter, *next;

  free_dictionary (agr_dict);
  for (iter = agr_first; iter; iter = next)
    {
      next = iter->next;

      if (iter->src->type == ALPHA)
	{
	  int i;

	  for (i = 0; i < agr_func_tab[iter->function].n_args; i++)
	    free (iter->arg[i].c);
	}
    }
}

/* Execution. */

static void accumulate_aggregate_info (ccase *input);
static void dump_aggregate_info (ccase *output);

/* Processes a single case INPUT for aggregation.  If output is
   warranted, it is written to case OUTPUT, which may be (but need not
   be) an alias to INPUT.  Returns -1 when output is performed, -2
   otherwise. */
/* The code in this function has an eerie similarity to
   vfm.c:SPLIT_FILE_procfunc()... */
static int
aggregate_single_case (ccase *input, ccase *output)
{
  static value *prev_case;

  /* The first case always begins a new break group.  We also need to
     preserve the values of the case for later comparison. */
  if (case_count++ == 0)
    {
      int n_elem = 0;
      
      {
	int i;

	for (i = 0; i < nv_sort; i++)
	  n_elem += v_sort[i]->nv;
      }
      
      if (prev_case)
	free (prev_case);
      prev_case = xmalloc (sizeof (value) * n_elem);

      /* Copy INPUT into prev_case. */
      {
	value *iter = prev_case;
	int i;

	for (i = 0; i < nv_sort; i++)
	  {
	    variable *v = v_sort[i];
	    
	    if (v->type == NUMERIC)
	      (iter++)->f = input->data[v->fv].f;
	    else
	      {
		/* PORTME: See TODO, portability note #1. */
		memcpy (iter->s, input->data[v->fv].s, v->width);
		iter += v->nv;
	      }
	  }
      }
	    
      accumulate_aggregate_info (input);
	
      return -2;
    }
      
  /* Compare the value of each break variable to the values on the
     previous case. */
  {
    value *iter = prev_case;
    int i;
    
    for (i = 0; i < nv_sort; i++)
      {
	variable *v = v_sort[i];
      
	switch (v->type)
	  {
	  case NUMERIC:
	    if (approx_ne (input->data[v->fv].f, iter->f))
	      goto not_equal;
	    iter++;
	    break;
	  case ALPHA:
	    /* PORTME: See TODO, portability note #1. */
	    if (memcmp (input->data[v->fv].s, iter->s, v->width))
	      goto not_equal;
	    iter += v->nv;
	    break;
	  default:
	    assert (0);
	  }
      }
  }

  accumulate_aggregate_info (input);

  return -2;
  
not_equal:
  /* The values of the break variable are different from the values on
     the previous case.  That means that it's time to dump aggregate
     info. */
  dump_aggregate_info (output);
  accumulate_aggregate_info (input);

  /* Copy INPUT into prev_case. */
  {
    value *iter = prev_case;
    int i;

    for (i = 0; i < nv_sort; i++)
      {
	variable *v = v_sort[i];
	    
	if (v->type == NUMERIC)
	  (iter++)->f = input->data[v->fv].f;
	else
	  {
	    /* PORTME: See TODO, portability note #1. */
	    memcpy (iter->s, input->data[v->fv].s, v->width);
	    iter += v->nv;
	  }
      }
  }
  
  return -1;
}

/* Accumulates aggregation data from the case INPUT. */
static void 
accumulate_aggregate_info (ccase *input)
{
  agr_var *iter;
  int weighting = (default_dict.weight_index != -1) << 6;

#define WEIGHT (input->data[default_dict.weight_index].f)

  for (iter = agr_first; iter; iter = iter->next)
    {
      value *v = &input->data[iter->src->fv];

      /* This is horrible.  There are too many possibilities. */
      switch (iter->function | ((iter->src->type==ALPHA) << 5) | weighting)
	{
	case SUM:
	case MEAN:
	  iter->dbl[0] += v->f;
	  break;
	case SUM | FWEIGHT:
	case MEAN | FWEIGHT:
	  iter->dbl[0] += v->f * WEIGHT;
	  break;
	case SD:
	  iter->dbl[0] += v->f;
	  iter->dbl[1] += v->f * v->f;
	  break;
	case SD | FWEIGHT:
	  {
	    double product = v->f * WEIGHT;
	    iter->dbl[0] += product;
	    iter->dbl[1] += product * v->f;
	    break;
	  }
	case MAX:
	case MAX | FWEIGHT:
	  iter->dbl[0] = max (iter->dbl[0], v->f);
	  break;
	case MAX | FSTRING:
	case MAX | FSTRING | FWEIGHT:
	  if (memcmp (iter->string, v->s, iter->src->width) < 0)
	    /* PORTME: See TODO, portability note #1. */
	    memcpy (iter->string, v->s, iter->src->width);
	  break;
	case MIN:
	case MIN | FWEIGHT:
	  iter->dbl[0] = min (iter->dbl[0], v->f);
	  break;
	case MIN | FSTRING:
	case MIN | FSTRING | FWEIGHT:
	  if (memcmp (iter->string, v->s, iter->src->width) > 0)
	    /* PORTME: See TODO, portability note #1. */
	    memcpy (iter->string, v->s, iter->src->width);
	  break;
	case FGT:
	case PGT:
	  if (approx_gt (v->f, iter->arg[0].f))
	    iter->int1++;
	  iter->int2++;
	  break;
	case FGT | FWEIGHT:
	case PGT | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (approx_gt (v->f, iter->arg[0].f))
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	case FGT | FSTRING:
	case PGT | FSTRING:
	  if (memcmp (iter->arg[0].c, v->s, iter->src->width) < 0)
	    iter->int1++;
	  iter->int2++;
	  break;
	case FGT | FSTRING | FWEIGHT:
	case PGT | FSTRING | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (memcmp (iter->arg[0].c, v->s, iter->src->width) < 0)
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	case FLT:
	case PLT:
	  if (approx_lt (v->f, iter->arg[0].f))
	    iter->int1++;
	  iter->int2++;
	  break;
	case FLT | FWEIGHT:
	case PLT | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (approx_lt (v->f, iter->arg[0].f))
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	case FLT | FSTRING:
	case PLT | FSTRING:
	  if (memcmp (iter->arg[0].c, v->s, iter->src->width) > 0)
	    iter->int1++;
	  iter->int2++;
	  break;
	case FLT | FSTRING | FWEIGHT:
	case PLT | FSTRING | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (memcmp (iter->arg[0].c, v->s, iter->src->width) > 0)
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	case FIN:
	case PIN:
	  if (approx_in_range (v->f, iter->arg[0].f, iter->arg[1].f))
	    iter->int1++;
	  iter->int2++;
	  break;
	case FIN | FWEIGHT:
	case PIN | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (approx_in_range (v->f, iter->arg[0].f, iter->arg[1].f))
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	case FIN | FSTRING:
	case PIN | FSTRING:
	  if (memcmp (iter->arg[0].c, v->s, iter->src->width) <= 0
	      && memcmp (iter->arg[1].c, v->s, iter->src->width) >= 0)
	    iter->int1++;
	  iter->int2++;
	  break;
	case FIN | FSTRING | FWEIGHT:
	case PIN | FSTRING | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (memcmp (iter->arg[0].c, v->s, iter->src->width) <= 0
		&& memcmp (iter->arg[1].c, v->s, iter->src->width) >= 0)
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	case FOUT:
	case POUT:
	  if (!approx_in_range (v->f, iter->arg[0].f, iter->arg[1].f))
	    iter->int1++;
	  iter->int2++;
	  break;
	case FOUT | FWEIGHT:
	case POUT | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (!approx_in_range (v->f, iter->arg[0].f, iter->arg[1].f))
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	case FOUT | FSTRING:
	case POUT | FSTRING:
	  if (memcmp (iter->arg[0].c, v->s, iter->src->width) > 0
	      && memcmp (iter->arg[1].c, v->s, iter->src->width) < 0)
	    iter->int1++;
	  iter->int2++;
	  break;
	case FOUT | FSTRING | FWEIGHT:
	case POUT | FSTRING | FWEIGHT:
	  {
	    double w = WEIGHT;
	    if (memcmp (iter->arg[0].c, v->s, iter->src->width) > 0
		&& memcmp (iter->arg[1].c, v->s, iter->src->width) < 0)
	      iter->dbl[0] += w;
	    iter->dbl[1] += w;
	    break;
	  }
	}
    }
}

/* We've come to a record that differs from the previous in one or
   more of the break variables.  Make an output record from the
   accumulated statistics in the OUTPUT case. */
static void 
dump_aggregate_info (ccase *output)
{
}

/* Resets the state for all the aggregate functions. */
static void
initialize_aggregate_functions (void)
{
  agr_var *iter;

  for (iter = agr_first; iter; iter = iter->next)
    {
      int plain_function = iter->function & ~FWEIGHT;
      
      switch (plain_function)
	{
	case MIN:
	  iter->dbl[0] = DBL_MAX;
	  break;
	case MIN | FSTRING:
	  /* PORTME: See TODO, portability note #1. */
	  memset (iter->string, 255, iter->src->width);
	  break;
	case MAX:
	  iter->dbl[0] = -DBL_MAX;
	  break;
	case MAX | FSTRING:
	  /* PORTME: See TODO, portability note #1. */
	  memset (iter->string, 0, iter->src->width);
	  break;
	default:
	  iter->dbl[0] = iter->dbl[1] = 0.0;
	  iter->int1 = iter->int2 = 0;
	  break;
	}
    }
}

/* Debugging. */
#if DEBUGGING
/* Print out useful debugging information. */
static void
debug_print (int flags)
{
  printf ("AGGREGATE\n /OUTFILE=%s\n",
	  outfile ? fh_handle_filename (outfile) : "*");

  if (missing == COLUMNWISE)
    puts (" /MISSING=COLUMNWISE");

  if (flags & 2)
    puts (" /DOCUMENT");
  if (flags & 4)
    puts (" /PRESORTED");
  
  {
    int i;

    printf (" /BREAK=");
    for (i = 0; i < nv_sort; i++)
      printf ("%s(%c) ", v_sort[i]->name,
	      v_sort[i]->p.srt.order == SRT_ASCEND ? 'A' : 'D');
    putc ('\n', stdout);
  }
  
  {
    agr_var *iter;
    
    for (iter = agr_first; iter; iter = iter->next)
      {
	agr_func *f = &agr_func_tab[iter->function];
	
	printf (" /%s", iter->dest->name);
	if (iter->dest->label)
	  printf ("'%s'", iter->dest->label);
	printf ("=%s(%s", f->name, iter->src->name);
	if (f->n_args)
	  {
	    int i;
	    
	    for (i = 0; i < f->n_args; i++)
	      {
		putc (',', stdout);
		if (iter->src->type == NUMERIC)
		  printf ("%g", iter->arg[i].f);
		else
		  printf ("%.*s", iter->src->width, iter->arg[i].c);
	      }
	  }
	printf (")\n");
      }
  }
}

#endif /* DEBUGGING */
