/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#if !var_h
#define var_h 1

#include "avl.h"
#include "hash.h"
#include "common.h"
#include "file-handle.h"

#include <time.h>

/* Values. */

/* Describes one value, which is either a floating-point number or a
   short string. */
typedef union
  {
    /* A numeric value. */
    double f;

    /* A short-string value. */
    unsigned char s[MAX_SHORT_STRING];

    /* This member is used by data-in.c to return a string result,
       since it may need to return a long string.  As currently
       implemented, it's a pointer to a static internal buffer in
       data-in.c.

       Also used by evaluate_expression() to return a string result.
       As currently implemented, it's a pointer to a dynamic buffer in
       the appropriate expression.

       Also used by the AGGREGATE procedure in handling string
       values. */
    unsigned char *c;
  }
value;

/* Describes one value label. */
typedef struct
  {
    value v;			/* The value being labeled. */
    char *s;			/* Pointer to malloc()'d label. */
    int ref_count;		/* Reference count. */
  }
value_label;

/* Display formats. */

/* Display format types. */
/* See the definitions of these functions and variables when modifying
   this list:
   misc.c:formats[]
   misc.c:convert_fmt_ItoO()
   sfm-read.c:parse_format_spec()
   sfm-write.c:write_format_spec()
   data-in.c:parse_string_as_format()
   data-out.c:convert_format_to_string(). */
enum
  {
    /* Numeric and string formats. */
    FMT_F,
    FMT_N,
    FMT_E,
    FMT_COMMA,
    FMT_DOT,
    FMT_DOLLAR,
    FMT_PCT,
    FMT_Z,
    FMT_A,
    FMT_AHEX,
    FMT_IB,
    FMT_P,
    FMT_PIB,
    FMT_PIBHEX,
    FMT_PK,
    FMT_RB,
    FMT_RBHEX,

    /* Custom currency. */
    FMT_CCA,
    FMT_CCB,
    FMT_CCC,
    FMT_CCD,
    FMT_CCE,

    /* Date/time formats. */
    FMT_DATE,
    FMT_EDATE,
    FMT_SDATE,
    FMT_ADATE,
    FMT_JDATE,
    FMT_QYR,
    FMT_MOYR,
    FMT_WKYR,
    FMT_DATETIME,
    FMT_TIME,
    FMT_DTIME,
    FMT_WKDAY,
    FMT_MONTH,
    
    /* These aren't real formats.  They're used by DATA LIST. */
    FMT_T,
    FMT_X,
    FMT_DESCEND,
    FMT_NEWREC,
    
    FMT_NUMBER_OF_FORMATS,
  };

/* Describes one of the display formats above. */
typedef struct
  {
    char name[9];		/* `DATETIME' is the longest. */
    int n_args;			/* 1=width; 2=width.decimals. */
    int Imin_w, Imax_w;		/* Limits on input width. */
    int Omin_w, Omax_w;		/* Limits on output width. */
    int cat;			/* Category. */
    int output;			/* Output format. */
  }
fmt_desc;

/* Display format categories. */
enum
  {
    FCAT_BLANKS_SYSMIS = 001,	/* 1=All-whitespace means SYSMIS. */
    FCAT_EVEN_WIDTH = 002,	/* 1=Width must be even. */
    FCAT_STRING = 004,		/* 1=String input/output format. */
    FCAT_SHIFT_DECIMAL = 010,	/* 1=Automatically shift decimal point
				   on output--used for fixed-point
				   formats. */
    FCAT_OUTPUT_ONLY = 020,	/* 1=This is not an input format. */
  };

/* Display format. */
typedef struct fmt_spec
  {
    int type;			/* One of the above constants. */
    int w;			/* Width. */
    int d;			/* Number of implied decimal places. */
  }
fmt_spec;

/* Descriptions of all the display formats above. */
extern fmt_desc formats[];

/* Set this flag to cause errors to return SYSMIS but not report an
   error to the user. */
extern int fmt_parse_ignore_error;

int parse_format_specifier (fmt_spec *input, int allow_xt);
int parse_format_specifier_name (char **cp, int allow_xt);
int check_input_specifier (fmt_spec *spec);
int check_output_specifier (fmt_spec *spec);
int check_string_specifier (fmt_spec *spec, int min_len);
void convert_fmt_ItoO (fmt_spec *input, fmt_spec *output);
int parse_string_as_format (char *s, int len, fmt_spec *fp, int fc, value *v);
int convert_format_to_string (char *s, fmt_spec *fp, value *v);
char *fmt_to_string (fmt_spec *);
void num_to_string (double v, char *s, int w, int d);

/* Frequency tables. */

/* Frequency table entry. */
typedef struct
  {
    value v;			/* The value. */
    double c;			/* The number of occurrences of the value. */
  }
freq;

/* Types of frequency tables. */
enum
  {
    FRQM_GENERAL,
    FRQM_INTEGER,
  };

/* Entire frequency table. */
typedef struct
  {
    int mode;			/* FRQM_GENERAL or FRQM_INTEGER. */

    /* General mode. */
    avl_tree *tree;		/* Undifferentiated data. */

    /* Integer mode. */
    double *vector;		/* Frequencies proper. */
    int min, max;		/* The boundaries of the table. */
    double out_of_range;	/* Sum of weights of out-of-range values. */
    double sysmis;		/* Sum of weights of SYSMIS values. */

    /* All modes. */
    freq *valid;		/* Valid freqs. */
    int n_valid;		/* Number of total freqs. */

    freq *missing;		/* Missing freqs. */
    int n_missing;		/* Number of missing freqs. */

    /* Statistics. */
    double total_cases;		/* Sum of weights of all cases. */
    double valid_cases;		/* Sum of weights of valid cases. */
  }
freq_tab;

/* A complete set of 3 frequency tables. */
typedef struct
  {
    freq_tab miss;		/* Includes user-missing values. */
    freq_tab no_miss;		/* Excludes user-missing values. */
    freq_tab sel;		/* Identical to either miss or no_miss. */
  }
freq_tab_set;

/* Procedures' private per-variable data. */

/* Structure name suffixes for private data:
   _proc: for a procedure (i.e., LIST -> list_proc).
   _trns: for a transformation (i.e., COMPUTE -> compute_trns.
   _pgm: for an input program (i.e., DATA LIST -> data_list_pgm). */

typedef struct variable variable;

/* CROSSTABS private data. */
typedef struct
  {
    int selected;		/* Nonzero if specified on TABLES. */
    int min;			/* Minimum value, in integer mode. */
    int cur;			/* Currently selected value, in integer mode. */
    int max;			/* Maximum value, in integer mode. */
    freq *f;			/* Current value, in general mode. */
    avl_traverser *t;		/* For stepping f, in general mode. */
  }
crosstab_proc;

/* FREQUENCIES private data. */
enum
  {
    frq_mean = 0, frq_semean, frq_median, frq_mode, frq_stddev, frq_variance,
    frq_kurt, frq_sekurt, frq_skew, frq_seskew, frq_range, frq_min, frq_max,
    frq_sum, frq_n_stats,
  };

typedef struct
  {
    /* General mode. */
    freq_tab tab;		/* Frequencies table to use. */

    /* Percentiles. */
    int n_groups;		/* Number of groups. */
    double *groups;		/* Groups. */

    /* Statistics. */
    double stat[frq_n_stats];
  }
frequencies_proc;

/* LIST private data. */
typedef struct
  {
    int newline;		/* Whether a new line begins here. */
    int width;			/* Field width. */
    int vert;			/* Whether to print the varname vertically. */
  }
list_proc;

/* DESCRIPTIVES private data.  Note that the DESCRIPTIVES procedure also
   has a transformation, descriptives_trns. */
enum
  {
    /* As these are used as bit indexes, there must be 32 or fewer.
       Be very careful in adjusting these, see the structure below
       and the table in descriptives.q. */
    dsc_mean = 0, dsc_semean, dsc_stddev, dsc_variance, dsc_kurt,
    dsc_sekurt, dsc_skew, dsc_seskew, dsc_range, dsc_min,
    dsc_max, dsc_sum, dsc_n_stats,
  };

typedef struct
  {
    /* Miscellaneous. */
    int dup;			/* Finds duplicates in list of
				   variables. */
    char zname[10];		/* Name for z-score variable. */

    /* Counts. */
    double valid, miss;		/* Valid, missing--general. */

    /* Mean, moments about the mean. */
    double X_bar, M2, M3, M4;
    double min, max;

    /* Statistics. */
    double stats[dsc_n_stats];	/* Everything glommed together. */
  }
descriptives_proc;

/* GET private data. */
typedef struct
  {
    int fv, nv;			/* First, last, # of values. */
  }
get_proc;

/* Sort order. */
enum
  {
    SRT_ASCEND,			/* A, B, C, ..., X, Y, Z. */
    SRT_DESCEND,		/* Z, Y, X, ..., C, B, A. */
  };

/* SORT CASES private data. */
typedef struct
  {
    int order;			/* SRT_ASCEND or SRT_DESCEND. */
  }
sort_cases_proc;

/* MODIFY VARS private data. */
typedef struct
  {
    char new_name[9];		/* Variable's new name. */
    int drop_this_var;		/* 0=keep this var, 1=drop this var. */
    variable *next;		/* Next in linked list. */
  }
modify_vars_proc;

/* Script variables. */

/* Variable type. */
enum
  {
    NUMERIC,			/* A numeric variable. */
    ALPHA,			/* A string variable.  (STRING is pre-empted by tokens.h) */
  };

/* Types of missing values.  Order is significant, see
   mis-val.c:parse_numeric(), sfm-read.c:sfm_read_dictionary()
   sfm-write.c:sfm_write_dictionary(), sysfile-info.c:cmd_sysfile_info(). */
enum
  {
    MISSING_NONE,		/* No user-missing values. */
    MISSING_1,			/* One user-missing value. */
    MISSING_2,			/* Two user-missing values. */
    MISSING_3,			/* Three user-missing values. */
    MISSING_RANGE,		/* [a,b]. */
    MISSING_LOW,		/* (-inf,a]. */
    MISSING_HIGH,		/* (a,+inf]. */
    MISSING_RANGE_1,		/* [a,b], c. */
    MISSING_LOW_1,		/* (-inf,a], b. */
    MISSING_HIGH_1,		/* (a,+inf), b. */
  };

/* A variable's dictionary entry.  Note: don't reorder name[] from the
   first element; a pointer to `variable' should be a pointer to
   member `name'. */
struct variable
  {
    /* Very basic stuff. */
    char name[9];		/* As a string. */
    int type;			/* NUMERIC or ALPHA. */
    int width;			/* Size of string variables in chars. */
    int fv, nv;			/* Index into `value's, number of values. */
    int left;			/* 0=do not LEAVE, 1=LEAVE. */

    /* Indexes. */
    int index;			/* Index into var[]. */
    int prv_index;		/* Index used by various transformations. */

    /* Missing values. */
    int miss_type;		/* One of the MISSING_* constants. */
    value missing[3];		/* User-missing value. */

    /* Display formats. */
    fmt_spec print;		/* Default format for PRINT. */
    fmt_spec write;		/* Default format for WRITE. */

    /* Labels. */
    avl_tree *val_lab;		/* Avltree of value_label structures. */
    char *label;		/* Variable label. */

    /* Per-procedure info. */
    union
      {
	crosstab_proc crs;
	descriptives_proc dsc;
	frequencies_proc frq;
	list_proc lst;
	get_proc get;
	sort_cases_proc srt;
	modify_vars_proc mfv;
      }
    p;
  }
/* variable */ ;

/* Represents a single variable on SPLIT FILE. */
typedef struct
  {
    /* The variable. */
    variable *v;

    /* The variable's data. */
    int type;			/* NUMERIC, ALPHA, or -1 as sentinel. */
    int width;			/* (string) Width. */
    int fv;			/* First `value' index. */
  }
split;

/* Represents a vector as created by the VECTOR transformation. */
typedef struct
  {
    int index;			/* Index into vec[]. */
    char name[9];		/* Name. */
    variable **v;		/* Vector of variables. */
    int nv;			/* Number of variables. */
  }
vector;

/* Cases. */

/* A single case. (This doesn't need to be a struct anymore, but it
   remains so for hysterical raisins.) */
typedef struct
  {
    value data[1];
  }
ccase;

/* Dictionary. */ 

/* Complete dictionary state. */
typedef struct
  {
    variable **var;		/* Variable descriptions. */
    avl_tree *var_by_name;	/* Variables arranged by name. */
    int nvar;			/* Number of variables. */

    int N;			/* Current case limit (N command). */

    int nval;			/* Number of value structures per case. */

    int n_splits;		/* Number of variables to split on. */
    split *splits;		/* Variables split on; n_splits+1 elements. */

    char *label;		/* File label. */

    int n_documents;		/* Number of lines of documents. */
    char *documents;		/* Documents; 80*n_documents bytes in size. */

    int weight_index;		/* `value' index of $WEIGHT, or -1 if none.
				   Call update_weighting() before using! */
    char weight_var[9];		/* Name of WEIGHT variable. */

    char filter_var[9];		/* Name of FILTER variable. */
    /* Do not make another field the last field! or see
       temporary.c:restore_dictionary() before doing so! */
  }
dictionary;

/* This is the active file dictionary. */
extern dictionary default_dict;

/* Transformation state. */

/* Default file handle for DATA LIST, REREAD, REPEATING DATA
   commands. */
extern file_handle *default_handle;

/* VECTOR vectors. */
extern vector *vec;
extern int nvec;

/* PROCESS IF expression. */
extern struct expression_struct *process_if_expr;

/* TEMPORARY support. */

/* 1=TEMPORARY has been executed at some point. */
extern int temporary;

/* If temporary!=0, the saved dictionary. */
extern dictionary *temp_dict;

/* If temporary!=0, index into t_trns[] (declared far below) that
   gives the point at which data should be written out.  -1 means that
   the data shouldn't be changed since all transformations are
   temporary. */
extern int temp_trns;

void cancel_temporary (void);

/* Functions. */

int is_varname (char *);
int is_dict_varname (dictionary *, char *);

/* Flags for passing to fill_all_vars(). */
enum
  {
    FV_NONE = 0,		/* No flags. */
    FV_NO_SYSTEM = 001,		/* Don't include system variables. */
    FV_NO_SCRATCH = 002,	/* Don't include scratch variables. */
  };

void fill_all_vars (variable ***, int *, int flags);

int val_lab_cmp (void *, void *, void *);
char *get_val_lab (variable *, value, int);
void free_val_lab (void *, void *);
void free_value_label (value_label *);

int alloc_val (int);
void dump_split_vars (ccase *);

int is_num_user_missing (double, variable *);
int is_str_user_missing (unsigned char[], variable *);
int is_missing (value *, variable *);
int is_user_missing (value *, variable *);

int cmp_variable (void *, void *, void *);

variable *create_variable (dictionary *, char *name, int type, int width);
void delete_variable (variable * v);
variable *find_variable (char *name);
variable *find_dict_variable (dictionary *, char *name);
void init_variable (dictionary *, variable *, char *name, int type, int width);
void replace_variable (variable *, char *name, int type, int width);
void clear_variable (dictionary *, variable *);
void rename_variable (dictionary *, variable * v, char *new_name);
void discard_variables (void);
void copy_variable (variable *dest, variable *src);
variable *dup_variable (dictionary *dict, variable *src, char *name);

variable *update_weighting (dictionary *);
void stop_weighting (dictionary *);

vector *find_vector (char *name);

dictionary *save_dictionary (void);
void restore_dictionary (dictionary *);
void free_dictionary (dictionary *);
dictionary *new_dictionary (void);

/* INPUT PROGRAM private data. */

/* Forward declarations. */
typedef struct trns_header_struct any_trns;

/* Bitmasks to indicate variable type. */
#define INP_MASK	3	/* 2#11 */
#define INP_NUMERIC	0	/* Numeric. */
#define INP_STRING	1	/* String. */
#define INP_RIGHT	0	/* Ordinary. */
#define INP_LEFT	2	/* Scratch or LEAVE. */

/* A bit-vector of two-bit entries.  The array tells INPUT PROGRAM how
   to initialize each `value'.  Modified by envector(), devector(), which
   are called by create_variable(), also by LEAVE, COMPUTE(!).  */
extern unsigned char *inp_init;

/* Number of bytes allocated for inp_init. */
extern size_t inp_init_size;

/* Transformations. */

/* Header for all transformations. */
typedef struct trns_header_struct
  {
    int index;			/* Index into t_trns[]. */
    int (*proc) (any_trns *, ccase *);	/* Transformation proc. */
    void (*free) (any_trns *);	/* Garbage collector proc. */
  }
trns_header;

/* Array of transformations */
extern any_trns **t_trns;

/* Number of transformations, maximum number in array currently. */
extern int n_trns, m_trns;

/* Index of first transformation that is really a transformation.  Any
   transformations before this belong to INPUT PROGRAM. */
extern int f_trns;

void add_transformation (any_trns * trns);
void cancel_transformations (void);

/* DESCRIPTIVES private data. */

/* Describes properties of a distribution for the purpose of
   calculating a Z-score. */
typedef struct
  {
    variable *s, *d;		/* Source, destination variable. */
    double mean;		/* Distribution mean. */
    double std_dev;		/* Distribution standard deviation. */
  }
dsc_z_score;

/* DESCRIPTIVES transformation (for calculating Z-scores). */
typedef struct
  {
    trns_header h;
    int n;			/* Number of Z-scores. */
    dsc_z_score *z;		/* Array of Z-scores. */
  }
descriptives_trns;

/* Vectors. */

/* A vector of longs. */
typedef struct
  {
    long *vec;			/* Contents. */
    int n;			/* Number of elements. */
    int m;			/* Number of elements room is allocated for. */
  }
long_vec;

/* cases.c */
void vec_init (long_vec *);
void vec_clear (long_vec *);
void vec_insert (long_vec *, long);
void vec_delete (long_vec *, long a, long b);
void devector (variable *);
void envector (variable *);

/* SORT CASES programmatic interface. */
int sort_cases (int separate);
void read_sort_output (int (*write_case)(void));

/* Variables to sort. */
extern variable **v_sort;
extern int nv_sort;

/* Virtual File Manager. */

/* This is the time at which vfm was last invoked. */
extern time_t last_vfm_invocation;

/* This is the case that is to be filled in by input programs. */
extern ccase *temp_case;

/* `value' indexes to initialize to particular values for certain cases. */
extern long_vec reinit_sysmis;	/* SYSMIS for every case. */
extern long_vec reinit_blanks;	/* Blanks for every case. */
extern long_vec init_zero;	/* Zero for first case only. */
extern long_vec init_blanks;	/* Blanks for first case only. */

/* A case stream: either a source or a sink, depending on context. */
typedef struct
  {
    /* Initializes sink. */
    void (*init) (void);
    
    /* Reads all the cases and passes them to write_case(). */
    void (*read) (void);

    /* Writes a single case, temp_case. */
    void (*write) (void);

    /* Switches mode from sink to source. */
    void (*mode) (void);
    
    /* Discards source's internal data. */
    void (*destroy_source) (void);

    /* Discards sink's internal data. */
    void (*destroy_sink) (void);

    /* Identifying name for the stream. */
    char *name;
  }
case_stream;

/* This is used to read from the active file. */
extern case_stream *vfm_source;

/* This is used to write to the replacement active file. */
extern case_stream *vfm_sink;

/* General data streams. */
extern case_stream vfm_memory_stream;
extern case_stream vfm_disk_stream;
extern case_stream sort_stream;

/* Streams that are only sources. */
extern case_stream data_list_source;
extern case_stream input_program_source;
extern case_stream file_type_source;
extern case_stream get_source;

/* Number of cases to lag. */
extern int n_lag;

void procedure (void (*beginfunc) (void),
		int (*procfunc) (ccase * curcase),
		void (*endfunc) (void));
int write_case (void);
ccase *lagged_case (int n_before);
void compact_case (ccase *dest, ccase *src);
void page_to_disk (void);

/* Variable parsers. */

/* Only parse_variables() supports options other than PV_APPEND,
   PV_SINGLE. */
typedef int pv_opts;		/* Used for PV_* constants. */
enum
  {
    PV_NONE = 0,		/* No options. */
    PV_SINGLE = 0001,		/* Restrict to a single varname or TO use. */
    PV_DUPLICATE = 0002,	/* Don't merge duplicates. */
    PV_APPEND = 0004,		/* Append to existing list. */
    PV_NO_DUPLICATE = 0010,	/* Error on duplicates. */
    PV_NUMERIC = 0020,		/* Vars must be numeric. */
    PV_STRING = 0040,		/* Vars must be string. */
    PV_SAME_TYPE = 00100,	/* All vars must be the same type. */
    PV_NO_SCRATCH = 00200,	/* Disallow scratch variables. */
  };

variable *parse_variable (void);
variable *parse_dict_variable (dictionary *);
int parse_variables (dictionary * dict, variable *** v, int *nv, pv_opts);
int parse_DATA_LIST_vars (char ***names, int *nnames, pv_opts);
int parse_mixed_vars (char ***names, int *nnames, pv_opts);

/* Settings. */

/* Current program state. */
enum
  {
    STATE_INIT,			/* initialization state */
    STATE_INPUT,		/* input state */
    STATE_TRANS,		/* transformation state */
    STATE_PROC,			/* procedure state */
    STATE_ERROR,		/* flags bad state transition */
  };
extern int pgm_state;

/* Borland C++, in another bow to inferiority, insists that any
   structure that is referenced, even as a pointer (like `struct
   expression_struct *') *must* be defined somewhere in the file.
   So we always include expr.h for BC++. */
#if __BORLANDC__
#include "expr.h"
#endif

#endif /* !var_h */
